import unittest
import asyncio
import tempfile
from carp.channel import UnixSocketChannel
from carp.host.host import Host, HostAnnounce
from carp.serializer import Serializer

from unittest import IsolatedAsyncioTestCase


class TestHostConnect(IsolatedAsyncioTestCase):

    def setUp(self):
        self.sockname = tempfile.mktemp()

    async def test_connect_back(self):
        """
        A pair of hosts, one in server mode, can make a connection
        """
        server_conn = asyncio.Event()
        client_conn = asyncio.Event()

        async def accept(event, channel):
            server_conn.set()

        async def connect(event, channel):
            client_conn.set()

        server_channel = UnixSocketChannel(socket_path=self.sockname)
        server_host = Host()
        server_host.on("accept", accept)

        client_channel = UnixSocketChannel(socket_path=self.sockname)
        client_host = Host()
        client_host.on("connect", connect)

        await server_host.start(server_channel)
        await client_host.connect(client_channel)
        await server_conn.wait()
        await client_conn.wait()

        await client_host.stop()
        await server_host.stop()

    async def test_message(self):
        """
        A pair of hosts, one in server mode, can make a connection
        and send a message
        """
        message_recvd = asyncio.Event()
        messages = []
        async def message(event, msg):
            messages.append(msg)
            message_recvd.set()

        server_channel = UnixSocketChannel(socket_path=self.sockname)
        server_host = Host()
        server_host.on("message", message)

        client_channel = UnixSocketChannel(socket_path=self.sockname)
        client_host = Host()

        await server_host.start(server_channel)
        await client_host.connect(client_channel)

        message = 'hello, world'
        await client_channel.put(Serializer.serialize(message))
        await message_recvd.wait()
        await client_host.stop()
        await server_host.stop()

        self.assertTrue(isinstance(messages[0], HostAnnounce))
        self.assertEqual(messages[1:], [message])



import datetime
from unittest import TestCase

from carp.serializer import JsonSerializer

pretty = b"""json:{
    "a": 1,
    "b": 2,
    "c": "hello",
    "__type__": "dict"
}"""

class TestJsonSerializer(TestCase):

    def test_serialize__no_options(self):
        """
        JSON serializer works for basic object
        """
        js = JsonSerializer()
        serialized = js.serialize(dict(a=1, b=2, c="hello"))
        self.assertEqual(
            serialized,
            b'json:{"a": 1, "b": 2, "c": "hello", "__type__": "dict"}'
        )

    def test_serialize__pretty(self):
        """
        JSON serializer works for basic object with options passed in
        """
        js = JsonSerializer(indent=4)
        serialized = js.serialize(dict(a=1, b=2, c="hello"))
        self.assertEqual(serialized, pretty)


    def test_serialize__times(self):
        """
        JSON serializer works for datetime objects
        """
        now = datetime.datetime.utcnow()

        js = JsonSerializer()
        self.assertEqual(
            js.serialize(dict(ts=now)),
            f'json:{{"ts": {{"__datetime__": true, "value": "{now.isoformat()}"}}, "__type__": "dict"}}'.encode('utf-8')
        )
        self.assertEqual(
            js.serialize(dict(ts=now.date())),
            f'json:{{"ts": {{"__date__": true, "value": "{now.date().isoformat()}"}}, "__type__": "dict"}}'.encode('utf-8')
        )

    def test_serialize__tuple(self):
        """
        JSON serializer works for tuple
        """
        js = JsonSerializer()
        serialized = js.serialize(dict(a=(1, 2, 3)))
        self.assertEqual(
            serialized,
            b'json:{"a": [1, 2, 3], "__type__": "dict"}'
        )

    def test_deserialize__basic(self):
        """
        JSON deserializer works for basic object
        """
        js = JsonSerializer()
        obj = js.deserialize(
            b'{"a": 1, "b": 2, "c": "hello"}'
        )
        self.assertEqual(
            obj,
            dict(a=1, b=2, c="hello")
        )



import asyncio
import tempfile
from carp.channel import UnixSocketChannel
from carp.service import ApiFunction
from carp.host import Host

from unittest import IsolatedAsyncioTestCase


class TestHostAnnounce(IsolatedAsyncioTestCase):

    def setUp(self):
        self.sockname = tempfile.mktemp()

    async def test_client_export_service(self):
        """
        A pair of hosts, one in server mode, can make a connection
        """
        server_channel = UnixSocketChannel(socket_path=self.sockname)
        server_host = Host()

        client_channel = UnixSocketChannel(socket_path=self.sockname)
        client_host = Host()

        await server_host.start(server_channel)
        await client_host.connect(client_channel)

        func_called = asyncio.Event()

        async def svc_func():
            func_called.set()
            return True

        await client_host.export(svc_func)
        clnt_svc = await server_host.require(svc_func)

        self.assertIn(clnt_svc.name, client_host.services_local)
        self.assertIn(clnt_svc.name, server_host.services_remote)
        self.assertEqual(
            server_host.services_remote[clnt_svc.name][0],
            client_host.id
        )
        await client_host.stop()
        await server_host.stop()

    async def test_server_export_service(self):
        """
        A pair of hosts, one in server mode, can make a connection
        """
        server_channel = UnixSocketChannel(socket_path=self.sockname)
        server_host = Host()

        client_channel = UnixSocketChannel(socket_path=self.sockname)
        client_host = Host()

        await server_host.start(server_channel)
        await client_host.connect(client_channel)

        func_called = asyncio.Event()

        async def svc_func():
            func_called.set()
            return True

        await server_host.export(svc_func)
        clnt_svc = await client_host.require(svc_func)

        self.assertIn(clnt_svc.name, server_host.services_local)
        self.assertIn(clnt_svc.name, client_host.services_remote)
        self.assertEqual(
            client_host.services_remote[clnt_svc.name][0],
            server_host.id
        )
        await client_host.stop()
        await server_host.stop()






import asyncio
from carp.serializer import Serializable, ProtobufSerializer
from carp.serializer.protobuf import call_data_pb2


class CallData(Serializable):
    """
    CallData is the generic wrapper for a remote method or function
    call. It packages up the args and the identity of the called
    object and provides an asyncio.Event() which will be notified when
    the remote call is complete.
    """
    serializer = ProtobufSerializer
    protobuf_type = call_data_pb2.CallData
    call_counter = 0

    def __init__(self, *, call_id=None, instance_id=None, service_name, host_id, args, kwargs):
        if call_id is None:
            self.call_id = CallData.call_counter
            CallData.call_counter += 1
        else:
            self.call_id = call_id

        self.service_name = service_name
        self.host_id = host_id
        self.instance_id = instance_id
        self.args = args or []
        self.kwargs = kwargs or {}
        self.event = asyncio.Event()
        self.response = None


class CallResponse(Serializable):
    """
    CallResponse is what you get back when the call is complete
    """
    serializer = ProtobufSerializer
    protobuf_type = call_data_pb2.CallResponse

    def __init__(self, *, call_id, service_name, host_id, value, exception):
        self.call_id = call_id
        self.service_name = service_name
        self.host_id = host_id
        self.value = value
        self.exception = exception

    def to_dict(self):
        return dict(
            call_id=self.call_id,
            service_name=self.service_name,
            host_id=self.host_id,
            value=self.value,
            exception=self.exception,
        )



import asyncio
from asyncio import IncompleteReadError

from .channel import Channel
from .exceptions import ConnectionStatusError, ChannelError


class UnixSocketChannel(Channel):
    SYNC_MAGIC = b"[ SYNC ]"
    MAX_MESSAGE_SIZE = 65535

    def __init__(self, *, socket_path: str):
        self.socket_path = socket_path
        self.status = Channel.CLOSED
        self.reader = None
        self.writer = None
        self.server = None

    async def connect(self):
        self.status = Channel.CONNECTING
        reader, writer = await asyncio.open_unix_connection(self.socket_path)
        self.reader = reader
        self.writer = writer
        self.status = Channel.CONNECTED

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
            self.writer = None
        if self.server:
            self.server.close()
            await self.server.wait_closed()
            self.server = None
        self.status = Channel.CLOSED

    async def serve(self, *, on_connect):
        async def _connected_cb(reader, writer):
            connected = UnixSocketChannel(socket_path=self.socket_path)
            connected.reader = reader
            connected.writer = writer
            connected.status = Channel.CONNECTED
            if asyncio.iscoroutinefunction(on_connect):
                await on_connect(connected)
            else:
                on_connect(connected)
        self.status = Channel.SERVING
        self.server = await asyncio.start_unix_server(
            _connected_cb, path=self.socket_path
        )

    async def put(self, message: bytes):
        if self.status != Channel.CONNECTED:
            raise ConnectionStatusError
        try:
            self.writer.write(self.SYNC_MAGIC)
            self.writer.write(b"% 8d" % len(message))
            self.writer.write(message)
            await self.writer.drain()
        except Exception as e:
            self.status = Channel.CLOSED
            raise ConnectionError("Connection closed during write")

    async def get(self):
        magiclen = len(self.SYNC_MAGIC)
        msglength = None
        badlength = False
        try:
            syncblock = await self.reader.readexactly(magiclen + 8)
            while msglength is None:
                if badlength or syncblock[:magiclen] != self.SYNC_MAGIC:
                    syncblock = syncblock[1:] + await self.reader.readexactly(1)
                    badlength = False
                    continue
                try:
                    msglength = int(syncblock[magiclen:])
                except ValueError:
                    badlength = True
                    continue

                if msglength < 0 or msglength > self.MAX_MESSAGE_SIZE:
                    badlength = True
                    msglength = None

            return await self.reader.readexactly(msglength)

        except IncompleteReadError as e:
            self.status = Channel.CLOSED
            raise ChannelError("Connection closed during read")



from carp.serializer import Serializable, Serializer
from carp.serializer.protobuf.compiled_python.envelope_pb2 import Envelope
from carp.serializer.protobuf.compiled_python.pytypes_pb2 import (
    PythonDict,
    PythonArray,
    PythonValue
)

SCALAR_TYPES = (int, float, str, bool)


class ProtobufSerializer(Serializer):
    label = "pb2"

    def __init__(self, **serialize_options):
        self.serialize_options = serialize_options
        super().__init__()

    def _pb2_value(self, val, pb=None):
        if not pb:
            pb = PythonValue()
        if isinstance(val, Serializable):
            pb._serialized = val.serialize()
        elif isinstance(val, str):
            pb._string = val
        elif isinstance(val, bytes):
            pb._bytes = val
        elif isinstance(val, bool):
            pb._bool = val
        elif isinstance(val, int):
            pb._int = val
        elif isinstance(val, float):
            pb._double = val
        elif isinstance(val, type(None)):
            pb._none = True
        elif isinstance(val, dict):
            pbdict = PythonDict()
            for key, value in val.items():
                pbitem = pbdict.items.add()
                pbitem.key.CopyFrom(self._pb2_value(key))
                pbitem.value.CopyFrom(self._pb2_value(value))
            pb._dict.CopyFrom(pbdict)
        elif isinstance(val, (list, tuple)):
            pbarray = PythonArray()
            for v in val:
                pbval = pbarray.items.add()
                pbval = self._pb2_value(v, pb=pbval)
            pb._array.CopyFrom(pbarray)
        return pb

    def _python_value(self, pb_val):
        if isinstance(pb_val, PythonArray):
            return [
                self._python_value(v) for v in pb_val.items
            ]
        elif isinstance(pb_val, PythonDict):
            return {
                self._python_value(i.key): self._python_value(i.value)
                for i in pb_val.items
            }
        elif isinstance(pb_val, PythonValue):
            py_type = pb_val.WhichOneof("value_types")
            if py_type == "_serialized":
                return Serializer.deserialize(pb_val._serialized)
            elif py_type == "_array":
                return self._python_value(pb_val._array)
            elif py_type == "_dict":
                return self._python_value(pb_val._dict)
            elif py_type == "_bool":
                return bool(getattr(pb_val, py_type))
            elif py_type == "_none":
                return None
            elif py_type:
                return getattr(pb_val, py_type)
            return None
        else:
            return pb_val

    def serialize(self, pyobj):
        obj_pb = pyobj.protobuf_type()
        obj_dict = {
            field.name: getattr(pyobj, field.name)
            for field in obj_pb.DESCRIPTOR.fields
        }
        pb_fields = {
            field.name: field
            for field in obj_pb.DESCRIPTOR.fields
        }
        obj_typename = type(pyobj).__name__

        for key, value in obj_dict.items():
            field_desc = pb_fields[key]
            if field_desc.message_type and field_desc.message_type.name == 'PythonValue':
                getattr(obj_pb, key).CopyFrom(self._pb2_value(value))
            elif isinstance(value, Serializable):
                val_pb = PythonValue()
                val_pb._serialized = value.serialize()
                setattr(obj_pb, key, val_pb)
            elif isinstance(value, SCALAR_TYPES):
                setattr(obj_pb, key, value)
            elif isinstance(value, dict):
                dict_pb = getattr(obj_pb, key)
                for dkey, dval in value.items():
                    ditem = dict_pb.items.add(
                        key=self._pb2_value(dkey),
                        value=self._pb2_value(dval)
                    )
            elif isinstance(value, (list, tuple)):
                array_pb = getattr(obj_pb, key)
                for dval in value:
                    ditem = array_pb.items.add()
                    ditem.CopyFrom(self._pb2_value(dval))

        envelope = Envelope()
        envelope.content_type = obj_typename
        envelope.content = obj_pb.SerializeToString()
        return f"{self.label}".encode('utf-8') + b':' + envelope.SerializeToString()

    def deserialize(self, bytestr):
        envelope = Envelope()
        envelope.ParseFromString(bytestr)

        # find the type object of the contents
        type_obj = Serializable.registry.get(envelope.content_type)
        if not type_obj:
            raise ValueError(f"No Serializable type for '{envelope.content_type}'")

        # deserialize the contents
        proto_obj = type_obj.protobuf_type()
        proto_obj.ParseFromString(envelope.content)

        obj_dict = {}
        for field in proto_obj.DESCRIPTOR.fields:
            raw_value = getattr(proto_obj, field.name)
            obj_dict[field.name] = self._python_value(raw_value)

        return type_obj(**obj_dict)





# build with 'python ./setup.py install'
from distutils.core import setup

VERSION = "0.0.2"

setup(
    name = 'carp-rpc',
    version = VERSION,
    license = 'MIT',
    description = 'Async RPC toolkit',
    author = 'Bill Gribble',
    author_email = 'grib@billgribble.com',
    url = 'https://github.com/bgribble/carp',
    download_url = 'https://github.com/bgribble/carp/archive/refs/tags/v0.0.2.tar.gz',
    keywords = ['rpc', 'protobuf', 'json'],
    install_requires = [
        "protobuf", "python-dateutil",
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
)



import unittest
import asyncio
import tempfile
from carp.channel import Channel, UnixSocketChannel

from unittest import IsolatedAsyncioTestCase

class TestUnixSocketChannel(IsolatedAsyncioTestCase):

    def setUp(self):
        self.sockname = tempfile.mktemp()

    async def test_serve(self):
        """
        A client+server pair can make a connection
        """
        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=lambda channel: None)

        self.assertEqual(server.status, Channel.SERVING)
        self.assertEqual(client.status, Channel.CLOSED)

        await client.connect()
        self.assertEqual(client.status, Channel.CONNECTED)

        await client.close()
        self.assertEqual(client.status, Channel.CLOSED)

        await server.close()
        self.assertEqual(server.status, Channel.CLOSED)

    async def test_client_message(self):
        """
        client can connect and send a message, which is received
        """
        message_recvd = asyncio.Event()
        messages = []

        async def on_connect(peer_channel):
            messages.append(await peer_channel.get())
            message_recvd.set()

        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=on_connect)
        await client.connect()
        await client.put(b"test message")
        
        await message_recvd.wait()
        self.assertEqual(messages[0], b"test message")

    async def test_server_message(self):
        """
        client can connect and send a message, which is received
        """
        async def on_connect(peer_channel):
            await peer_channel.put(b"test message")

        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=on_connect)
        await client.connect()
        message = await client.get() 
        self.assertEqual(message, b"test message")

    async def test_resync(self):
        """
        client can receive a message despite sync error 
        """
        async def on_connect(peer_channel):
            await peer_channel.put(b"test message 1")
            peer_channel.writer.write(b"XXX")
            await peer_channel.put(b"test message 2")

        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=on_connect)
        await client.connect()
        
        msg_1 = await client.get()
        msg_2 = await client.get()
        self.assertEqual(msg_1, b"test message 1")
        self.assertEqual(msg_2, b"test message 2")

    async def test_bad_message_length__not_number(self):
        """
        client can receive a message despite sync error 
        """
        async def on_connect(peer_channel):
            peer_channel.writer.write(peer_channel.SYNC_MAGIC)
            peer_channel.writer.write(b"XXX")
            await peer_channel.put(b"test message")

        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=on_connect)
        await client.connect()
        
        msg_1 = await client.get()
        self.assertEqual(msg_1, b"test message")

    async def test_bad_message_length__too_long(self):
        """
        client can receive a message despite sync error 
        """
        async def on_connect(peer_channel):
            peer_channel.writer.write(peer_channel.SYNC_MAGIC)
            peer_channel.writer.write(b'99999999') 
            peer_channel.writer.write(b'abc' * 100)
            await peer_channel.put(b"test message")

        server = UnixSocketChannel(socket_path=self.sockname)
        client = UnixSocketChannel(socket_path=self.sockname)

        await server.serve(on_connect=on_connect)
        await client.connect()
        
        msg_1 = await client.get()
        self.assertEqual(msg_1, b"test message")
